using System.Runtime.InteropServices;
using System.Collections;
using System.Collections.Generic;
using System.Numerics;
using System;
using System.Threading.Tasks;
using Microsoft.VisualBasic;

public class Seddona
{
    private readonly IntPtr _client;
    public ChainAPI Chain;
    public WalletAPI Wallet;

    public static Seddona GanacheClient()
    {
        return new Seddona(get_client(8545));
    }
    
    public static Seddona Client()
    {
        return new Seddona(get_client(8550));
    }
    
    public static Seddona Client(int port)
    {
        return new Seddona(get_client(port));
    }
   
    public Seddona() : this(get_client(8550)) {}
    
    public Seddona(IntPtr client)
    {
        _client = client;
        Chain = new ChainAPI(this);
        Wallet = new WalletAPI(this);
    }

    ~Seddona()
    {
        drop_client(_client);
    }

    public IntPtr ClientPtr()
    {
        return _client;
    }
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr get_client(int port);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr drop_client(IntPtr ptr);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern FFI.CallResultInternal get_address(IntPtr client);

    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr get_contract([MarshalAs(UnmanagedType.LPUTF8Str)] string address,
        [MarshalAs(UnmanagedType.LPUTF8Str)] string abi, IntPtr client);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr personal_sign([MarshalAs(UnmanagedType.LPUTF8Str)] string s, IntPtr client);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern void free_str(IntPtr s);
    
    
    public class WalletAPI
    {
        private readonly Seddona _client;
        public WalletAPI(Seddona client)
        {
            _client = client;
        }
        
        public async Task<string> GetAddress()
        {
            return await Task.Run(async () =>
            {
                var result = new FFI.CallResult(wallet_get_address(_client.ClientPtr()));
                IntPtr ptr = result.ResultPtr();
                var s = FFI.PtrToStringNoFree(ptr);
                return s;
            });
        }
        
        
        public async Task<BigInteger> GetBalance()
        {
            return await Task.Run(async () =>
            {
                var result = new FFI.CallResult(wallet_get_balance(_client.ClientPtr()));
                return BigInteger.Parse(FFI.PtrToStringNoFree(result.ResultPtr()));
            });
        }
       
        // TODO BigInt
        public async Task<ulong> GetTransactionCount()
        {
            return await Task.Run(async () =>
            {
                var result = new FFI.CallResult(wallet_get_transaction_count(_client.ClientPtr()));
                return (ulong) Marshal.ReadInt64(result.ResultPtr());
            });
        }
       
        // TODO
        public async Task<string> PersonalSign(string message)
        {
            return await Task.Run(async () =>
            {
                var result = new FFI.CallResult(wallet_personal_sign(message, _client.ClientPtr()));
                return FFI.PtrToStringNoFree(result.ResultPtr());
            });
        }
       
        // TODO add more functions to take in integers and floats and treat them accordingly
        public async Task<string> Transfer(string address, string amount)
        {
            if (address == "")
            {
                throw new Exception("Empty Address");
            }
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(wallet_transfer(address, amount, _client.ClientPtr()));
                return FFI.PtrToStringNoFree(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        static extern FFI.CallResultInternal wallet_get_address(IntPtr c);
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        static extern FFI.CallResultInternal wallet_get_balance(IntPtr c);
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        static extern FFI.CallResultInternal wallet_get_transaction_count(IntPtr c);
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        static extern FFI.CallResultInternal wallet_personal_sign([MarshalAs(UnmanagedType.LPUTF8Str)] string message, IntPtr c);
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        static extern FFI.CallResultInternal wallet_transfer([MarshalAs(UnmanagedType.LPUTF8Str)] string address, [MarshalAs(UnmanagedType.LPUTF8Str)] string amount, IntPtr c);
    }


    public class ChainAPI
    {
        private Seddona _client;
        public ChainAPI(Seddona client)
        {
            _client = client;
        }
        
        [Serializable]
        [StructLayout(LayoutKind.Sequential)]
        public partial class Block
        {
            public string hash;
            public Block() {}

            public string GetHash()
            {
                return hash;
            }
        }
        
        // TODO BigInt
        public async Task<UInt64> GetBlockHeight()
        {
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(chain_get_block_height(_client.ClientPtr()));
                return (UInt64) Marshal.ReadInt64(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        public static extern FFI.CallResultInternal chain_get_block_height(IntPtr c);
        
        public async Task<Block> GetLatestBlock()
        {
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(chain_get_latest_block(_client.ClientPtr()));
                return Marshal.PtrToStructure<Block>(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        public static extern FFI.CallResultInternal chain_get_latest_block(IntPtr c);
        
        public async Task<Block> GetBlockAtHeight(ulong height)
        {
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(chain_get_block_at_height(height, _client.ClientPtr()));
                return Marshal.PtrToStructure<Block>(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        public static extern FFI.CallResultInternal chain_get_block_at_height(ulong height, IntPtr c);

        [Serializable]
        [StructLayout(LayoutKind.Sequential)]
        public partial class Transaction
        {
            public uint BlockNumber;
            // public readonly uint TransactionType;
            public IntPtr Hash;
            public IntPtr From;
            public Transaction() {}

            ~Transaction()
            {
                FFI.free_str(Hash);
                FFI.free_str(From);
            }

            public string GetHash()
            {
                return FFI.PtrToStringNoFree(Hash);
            }
            
            public string GetFrom()
            {
                return FFI.PtrToStringNoFree(From);
            }
        }
        public async Task<Transaction> GetTransaction(string hash)
        {
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(chain_get_transaction(hash, _client.ClientPtr()));
                return Marshal.PtrToStructure<Transaction>(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        // returns transaction
        public static extern FFI.CallResultInternal chain_get_transaction([MarshalAs(UnmanagedType.LPUTF8Str)] string s, IntPtr c);

        [Serializable]
        [StructLayout(LayoutKind.Sequential)]
        public class Network
        {
            public readonly IntPtr Name;
            public readonly uint Id;

            public Network() {}

            ~Network()
            {
                FFI.free_str(Name);
            }

            public string GetName()
            {
                return FFI.PtrToStringNoFree(Name);
            }
        }

        public async Task<Network> GetNetwork()
        {
            return await Task.Run(() =>
            {
                var result = new FFI.CallResult(chain_get_network(_client.ClientPtr()));
                return Marshal.PtrToStructure<Network>(result.ResultPtr());
            });
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        // returns network
        public static extern FFI.CallResultInternal chain_get_network(IntPtr c); // 
        
        public async Task<ulong> GetGasPrice()
        {
            throw new NotImplementedException("TODO");
        }
        
        [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
        public static extern FFI.CallResultInternal chain_get_gas_price(); // was ulong
    }
}

// TODO serialize into 
public class Contract
{
    private readonly IntPtr _contract;
    private FFI.CallResult _cr;
    private readonly Seddona _client;
    
    public static Contract Deploy(string json, Seddona client)
    {
        var result = new FFI.CallResult(deploy_contract(json, client.ClientPtr()));
        return new Contract(result.ResultPtr(), result, client);
    }
    
    public Contract(string address, string abi, Seddona client)
    {
        var cr = new FFI.CallResult(get_contract(address, abi));
        _contract = cr.ResultPtr();
        _cr = cr; // hold onto to prevent cleanup
        _client = client;
    }

    ~Contract()
    {
    }
    
    public static async Task<Contract> DeployNftTokenFaucet(Seddona client)
    {
        return await Task.Run(() =>
        {
            var cr = new FFI.CallResult(deploy_nft_token_faucet(client.ClientPtr()));
            cr.SetTag("Deployed NFT Token");
            return new Contract(cr.ResultPtr(), cr, client);
        });
    }
    
    public Contract(IntPtr ptr, FFI.CallResult cr, Seddona client)
    {
        _contract = ptr;
        _cr = cr; 
        _client = client;
    }

    public string GetAddress()
    {
        var result = new FFI.CallResult(contract_get_address(_contract));
        return FFI.PtrToStringNoFree(result.ResultPtr());
    }

    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    // returns string
    private static extern FFI.CallResultInternal contract_get_address(IntPtr contract);
    
    /*
     * Execute function for contract on chain returns txid
     */
    // TODO txid should be its own type
    // TODO pass param list correctly
    public async Task<string> Call(string func, Abi.Token[] tokens)
    {
        return await Task.Run(() =>
        {
            var r = contract_call(func, _contract, tokens, tokens.Length, _client.ClientPtr());
            var result = new FFI.CallResult(r);
            return FFI.PtrToStringNoFree(result.ResultPtr());
        });
    }

    public async Task<string> Call(string func)
    {
        return await Call(func, new Abi.Token[] {});
    }
    
    public async Task<string> Call(string func, Abi.Token token)
    {
        return await Call(func, new []
        {
            token,
        });
    }
    
    public async Task<Abi.Token> Query(string func, Abi.Token token)
    {
        return await Query(func, new[] {token});
    }
    
    public async Task<Abi.Token> Query(string func)
    {
        return await Query(func, new Abi.Token[] {});
    }
    
    public async Task<Abi.Token> Query(string func, Abi.Token[] tokens)
    {
        return await Task.Run(async () =>
        {
            var cr = new FFI.CallResult(contract_query(func, _contract, tokens, tokens.Length, _client.ClientPtr()));
            return Marshal.PtrToStructure<Abi.Token>(cr.ResultPtr());
        });
    }
    
    public async Task<Abi.Token> Query2(string func)
    {
        return await Task.Run(async () =>
        {
            var cr = new FFI.CallResult(return_tokens(func, _contract, new Abi.Token[]{}, 0, _client.ClientPtr()));
            return Marshal.PtrToStructure<Abi.Token>(cr.ResultPtr());
        });
    }
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    // returns contract
    private static extern FFI.CallResultInternal get_contract([MarshalAs(UnmanagedType.LPUTF8Str)] string address, [MarshalAs(UnmanagedType.LPUTF8Str)] string abi);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    // returns contract
    private static extern FFI.CallResultInternal deploy_contract([MarshalAs(UnmanagedType.LPUTF8Str)] string json, IntPtr client);

    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    // returns contract
    private static extern FFI.CallResultInternal deploy_nft_token_faucet(IntPtr client);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    // returns string
    private static extern FFI.CallResultInternal contract_call([MarshalAs(UnmanagedType.LPUTF8Str)] string func, IntPtr contract, [In, Out] Abi.Token[] tokens, int tokensLen, IntPtr client);
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    private static extern FFI.CallResultInternal contract_query([MarshalAs(UnmanagedType.LPUTF8Str)] string func, IntPtr contract, [In, Out] Abi.Token[] tokens, int tokensLen, IntPtr client);
    
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    private static extern FFI.CallResultInternal return_tokens([MarshalAs(UnmanagedType.LPUTF8Str)] string func, IntPtr contract, [In, Out] Abi.Token[] tokens, int tokensLen, IntPtr client);
}

public class NFT {
    private Contract _contract;
    private Seddona _client;
    
    public NFT(string address, Seddona client)
    {
        _contract = new Contract(address, ABIS.ERC721, client);
        _client = client;
    }
   
    // ERC721
    public async Task Transfer(string from, string to, BigInteger tokenId)
    {
        await _contract.Call("transfer", new[]
        {
            Abi.Token.IntoAddress(from),
            Abi.Token.IntoAddress(to),
            Abi.Token.IntoUint(tokenId),
        });
    }

    public async Task<BigInteger> balanceOf(string owner)
    {
        var retToken = await _contract.Query("balanceOf", Abi.Token.IntoAddress(owner));
        return retToken.GetUInt().GetValueOrDefault();
    }

    public async Task<string> ownerOf(BigInteger tokenId)
    {
        var retToken = await _contract.Query("ownerOf", Abi.Token.IntoUint(tokenId));
        return retToken.GetAddress();
    }
    
    public async Task safeTransferFrom(string from, string to, BigInteger tokenId)
    {
        await _contract.Call("safeTransferFrom", new[]
        {
            Abi.Token.IntoAddress(from),
            Abi.Token.IntoAddress(to),
            Abi.Token.IntoUint(tokenId),
        });
    }
    
    public async Task transferFrom(string from, string to, BigInteger tokenId)
    {
        await _contract.Call("safeTransferFrom", new[]
        {
            Abi.Token.IntoAddress(from),
            Abi.Token.IntoAddress(to),
            Abi.Token.IntoUint(tokenId),
        });
    }

    public async Task approve(string approved, BigInteger tokenId)
    {
        await _contract.Call("approve", new[]
        {
            Abi.Token.IntoAddress(approved),
            Abi.Token.IntoUint(tokenId),
        });
    }

    public async Task setApprovalForAll(string operatorAddress, bool approved)
    {
        await _contract.Call("setApprovalForAll", new []
        {
            Abi.Token.IntoAddress(operatorAddress),
            Abi.Token.IntoBool(approved),
        });
    }

    public async Task<string> getApproved(BigInteger tokenId)
    {
        var retToken = await _contract.Query("getApproved", Abi.Token.IntoUint(tokenId));
        return retToken.GetAddress();
    }

    public async Task<bool> isApprovedForAll(string owner, string operatorAddress)
    {
        var retToken = await _contract.Query("isApprovedForAll", new[]
        {
            Abi.Token.IntoAddress(owner),
            Abi.Token.IntoAddress(operatorAddress),
        });
        return retToken.GetBool().GetValueOrDefault();
    }
    
    // ERC721Metadata
    public async Task<string> name()
    {
        var retToken = await _contract.Query("name");
        return retToken.GetString();
    }

    public async Task<string> symbol()
    {
        var retToken = await _contract.Query("name");
        return retToken.GetString();
    }

    public async Task<string> tokenURI(BigInteger tokenId)
    {
        var retToken = await _contract.Query("tokenURI", Abi.Token.IntoUint(tokenId));
        return retToken.GetString();
    }
    
    // ERC721Enumerable
    public async Task<BigInteger> totalSupply()
    {
        var retToken = await _contract.Query("totalSupply");
        return retToken.GetUInt().GetValueOrDefault();
    }

    public async Task<BigInteger> tokenByIndex(BigInteger index)
    {
        var retToken = await _contract.Query("tokenByIndex", Abi.Token.IntoUint(index));
        return retToken.GetUInt().GetValueOrDefault();
    }

    public async Task<BigInteger> tokenOfOwnerByIndex(string owner, BigInteger tokenId)
    {
        var retToken = await _contract.Query("tokenOfOwnerByIndex", new []
            {
                Abi.Token.IntoAddress(owner),
                Abi.Token.IntoUint(tokenId),
            });
        return retToken.GetUInt().GetValueOrDefault();
    }
}

public class ERC20
{
    private Contract _contract;
    private Seddona _client;
    
    public ERC20(string address, Seddona client)
    {
        _contract = new Contract(address, ABIS.ERC20, client);
        _client = client;
    }

    public string GetAddress()
    {
        return _contract.GetAddress();
    }
    
    
    // function name() public view returns (string)
    public async Task<string> name()
    {
        var retToken = await _contract.Query("name");
        return retToken.GetString();
    }
    
    // function symbol() public view returns (string)
    public async Task<string> symbol()
    {
        var retToken = await _contract.Query("symbol");
        return retToken.GetString();
    }
    
    // function decimals() public view returns (uint8)
    public async Task<BigInteger> decimals()
    {
        var retToken = await _contract.Query("decimals");
        return retToken.GetInt().GetValueOrDefault();
    }
    
    // function totalSupply() public view returns (uint256)
    public async Task<BigInteger> totalSupply()
    {
        var retToken = await _contract.Query("totalSupply");
        return retToken.GetUInt().GetValueOrDefault();
    }
    
    // function balanceOf(address _owner) public view returns (uint256 balance)
    public async Task<BigInteger> balanceOf(string owner)
    {
        var retToken = await _contract.Query("balanceOf", new []
        {
            Abi.Token.IntoAddress(owner),
        });
        return retToken.GetUInt().GetValueOrDefault();
    }
    
    // function transfer(address _to, uint256 _value) public returns (bool success)
    public async Task transfer(string to, BigInteger value)
    {
        var retToken = await _contract.Call("transfer", new []
        {
            Abi.Token.IntoAddress(to),
            Abi.Token.IntoUint(value),
        });
    }
    
    // function transferFrom(address _from, address _to, uint256 _value) public returns (bool success)
    public async Task transferFrom(string from, string to, BigInteger value)
    {
        await _contract.Call("transferFrom", new []
        {
            Abi.Token.IntoAddress(from),
            Abi.Token.IntoAddress(to),
            Abi.Token.IntoUint(value),
        });
    }
    
    // function approve(address _spender, uint256 _value) public returns (bool success)
    public async Task approve(string spender, BigInteger value)
    {
        await _contract.Call("approve", new []
        {
            Abi.Token.IntoAddress(spender),
            Abi.Token.IntoUint(value),
        });
    }
    
    // function allowance(address _owner, address _spender) public view returns (uint256 remaining)
    public async Task<BigInteger> allowance(string owner, string spender)
    {
        var retToken = await _contract.Query("allowance", new []
        {
            Abi.Token.IntoAddress(owner),
            Abi.Token.IntoAddress(spender),
        });

        return retToken.GetUInt().GetValueOrDefault();
    }
}

public class Abi
{
    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public partial struct TokensInternal
    {
        public uint len;
        public IntPtr data;
        // public IntPtr data2;
        // public IntPtr data3;
    }
    
    // TODO token should have an an outer wrapper to allow handling of pointers
    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public partial struct Tokens
    {
        // Number of elements.
        public UInt64 len;
        // Pointer to start of immutable data.
        public IntPtr data;
    }
   
    public partial struct Tokens : IEnumerable<Token>
    {
        public Tokens(Token[] tokens)
        {
            // TODO FREE
            this.data = GCHandle.Alloc(tokens, GCHandleType.Pinned).AddrOfPinnedObject();
            this.len = (ulong) tokens.Length;
        }
        public Tokens(GCHandle handle, ulong count)
        {
            this.data = handle.AddrOfPinnedObject();
            this.len = count;
        }
        public Tokens(IntPtr handle, ulong count)
        {
            this.data = handle;
            this.len = count;
        }

        public Token this[int i]
        {
            get
            {
                if (i >= Count) throw new IndexOutOfRangeException();
                var size = Marshal.SizeOf(typeof(Token));
                var ptr = new IntPtr(data.ToInt64() + i * size);
                return Marshal.PtrToStructure<Token>(ptr);
            }
        }
        public Token[] Copied
        {
            get
            {
                var rval = new Token[len];
                for (var i = 0; i < (int) len; i++) {
                    rval[i] = this[i];
                }
                return rval;
            }
        }
        public int Count => (int) len;
        public IEnumerator<Token> GetEnumerator()
        {
            for (var i = 0; i < (int)len; ++i)
            {
                yield return this[i];
            }
        }
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }
    }

    // TODO simple pointer likely better
    [Serializable]
    [StructLayout(LayoutKind.Explicit)]
    public partial struct TokenUnion
    {
        // TODO deallocate
        [FieldOffset(0)] 
        public IntPtr Ptr;
        
        [FieldOffset(0)] 
        public IntPtr Str;
        
        [FieldOffset(0)]
        public Int64 Int;
        
        [FieldOffset(0)]
        public UInt64 Uint;
       
        // CANNOT do bool here it wont like it for FFI
    }

    [Serializable]
    public enum TokenTag
    {
        Address,
        FixedBytes,
        Bytes,
        Int,
        Uint,
        Bool,
        String,
        FixedArray,
        Array,
        Tuple,
    }

    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public partial struct Token
    {
        public TokenTag _tag;
        public TokenUnion _union;
       
        public Token(TokenTag tag, TokenUnion union)
        {
            _tag = tag;
            _union = union;
        }

        public TokenTag Tag()
        {
            return _tag;
        }
        
        public string? GetAddress()
        {
            if (_tag == TokenTag.Address)
            {
                return FFI.PtrToString(_union.Ptr);
            }

            return null;
        }
        
        public BigInteger? GetInt()
        {
            if (_tag == TokenTag.Int)
            {
                var s = FFI.PtrToString(_union.Str);
                BigInteger? i = BigInteger.Parse(s);
                return i;
            }

            return null;
        }
        
        public BigInteger? GetUInt()
        {
            if (_tag == TokenTag.Uint)
            {
                var s = FFI.PtrToString(_union.Str);
                BigInteger? i = BigInteger.Parse(s);
                return i;
            }

            return null;
        }
        
        public bool? GetBool()
        {
            if (_tag == TokenTag.Bool)
            {
                return _union.Uint > 0;
            }

            return null;
        }
        
        public string? GetString()
        {
            if (_tag == TokenTag.String)
            {
                return FFI.PtrToStringNoFree(_union.Str);
            }

            return null;
        }

        public Token[] GetArray()
        {
            if (_tag == TokenTag.Array && _union.Ptr != IntPtr.Zero)
            {
                var tokens = Marshal.PtrToStructure<Tokens>(_union.Ptr);
                var tokenArr = new Token[tokens.len];
                for (var i = 0; i < tokenArr.Length; i++)
                {
                    tokenArr[i] = tokens[i];
                }

                return tokenArr;

            }

            return null;
        }

        public static Token IntoAddress(string address)
        {
            var s = Marshal.StringToCoTaskMemAnsi(address);
            return new Token(TokenTag.Address, new TokenUnion
            {
               Str = s,
            });
        }
        
        public static Token IntoInt(Int64 i)
        {
            return new Token(TokenTag.Int, new TokenUnion {
                Int = i
            });
        }
        
        public static Token IntoUint(BigInteger i)
        {
            var s = Marshal.StringToCoTaskMemAnsi(i.ToString());
            return new Token(TokenTag.Uint, new TokenUnion {
                Str = s,
            });
        }
        
        public static Token IntoBool(bool b)
        {
            uint i = 0;
            if (b)
            {
                i = 1;
            }
            return new Token(TokenTag.Bool, new TokenUnion {
                Uint = i,
            });
        }
        
        public static Token IntoString(string s)
        {
            return new Token(TokenTag.String, new TokenUnion {
                Str =  Marshal.StringToCoTaskMemAnsi(s),
            });
        }
    }
}

public static class FFI
{
    public const string NativeLib = "seddona_sdk";

    // public class SeddonaException : Exception
    // {
    //     public SeddonaException() : base() {}
    //     public SeddonaException(string message) : base(message) {}
    // }

    /*
     * Result Type
     */
    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public struct FakeResultBool
    {
        private bool passed;
        public bool GetPassed()
        {
            return passed;
        }
    }
    
    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public struct FakeResultString
    {
        private string message;
        public string GetMessage()
        {
            return message;
        }
    }
    
    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void CallResultCleanup(IntPtr handle);
   
    // TODO create another outer class to wrap this inner struct 
    // allows for destructor when variable is no longer used
    [Serializable]
    [StructLayout(LayoutKind.Sequential)]
    public struct CallResultInternal
    {
        public IntPtr result;
        public int error;
        public IntPtr error_str;
        public CallResultCleanup cleanup;
        public IntPtr msg;
    }

    public class CallResult
    {
        private CallResultInternal _callResult;
        private string _errorMsg;
        private string _tag;

        public CallResult(CallResultInternal cr)
        {
            _callResult = cr;
            if (cr.msg != IntPtr.Zero)
            {
                Console.WriteLine("Rust Message");
                Console.WriteLine(FFI.PtrToString(cr.msg));
            }
            if (cr.error_str != IntPtr.Zero)
            {
                _errorMsg = FFI.PtrToString(cr.error_str);
            }
            else
            {
                _errorMsg = "";
            }
        }

        public void SetTag(string tag)
        {
            _tag = tag;
        }
        
        public bool IsError()
        {
            return _callResult.error != 0;
        }

        // TODO expose error codes to C# 
        public string ErrorMessage()
        {
            if (_callResult.error > 0)
            {
                if (_callResult.error >= ErrorEnum.Length)
                {
                    return "unknown error code: " + _callResult.error;
                }
                else
                {
                    var msg = ErrorEnum[_callResult.error] + "(" + _callResult.error + ") Error";
                    if (_errorMsg != "")
                    {
                        msg += ": " + _errorMsg;
                    }
                    return msg;
                }
            }
            else
            {
                return "no error";
            }
        }

        // public T Result()
        // {
        //     if (_callResult.result != IntPtr.Zero)
        //     {
        //         return Marshal.PtrToStructure<T>(_callResult.result);
        //     }
        //     if (IsError())
        //     {
        //         throw new Exception(ErrorMessage());
        //     } 
        //     throw new Exception("Null Pointer Error");
        // }

        public IntPtr ResultPtr()
        {
            if (_callResult.result != IntPtr.Zero)
            {
                return _callResult.result;
            }
            if (IsError())
            {
                throw new Exception(ErrorMessage());
            }
            throw new Exception("Null Pointer Error");
        }

        private static string[] ErrorEnum = new[]
        {
            "Ok",
            "Internal",
            "NullPointer",
            "ClientSetupError",
            "Web3ApiTransport",
            "Rpc",
            "Parse",
            "InsufficientFunds",
            "ContractQuery",
            "Abi",
            "GasEstimation",
            "ParseAddress",
            "ParseContract",
            "XYZ",
            
            "AbiInvalidName",
            "AbiInvalidData",
            "AbiParse",
            "AbiParseInt",
            "AbiHex",
            "AbiOther",
            
            "AbiInvalidOutputType",
            "Deployment",
            "AbiInterfaceUnsupported",
            
            "ApiUnreachable",
            "ApiDecoder",
            "ApiInvalidResponse",
            "ApiTransport",
            "ApiRpc",
            "ApiIo",
            "ApiRecovery",
            "ApiInternal",
        };
        
        ~CallResult()
        {
            _callResult.cleanup(_callResult.result);
        }
    }
    
    // should return the result
    public static string GetResultStringFfiSuccess()
    {
        var result = new CallResult(get_result_string_ffi_success());
        return PtrToStringNoFree(result.ResultPtr());
    }
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern CallResultInternal get_result_string_ffi_success();
    
    // should return the result type of FakeResult
    public static FakeResultBool GetResultStructBoolFfiSuccess()
    {
        var result = new CallResult(get_result_struct_bool_ffi_success());
        return Marshal.PtrToStructure<FakeResultBool>(result.ResultPtr());
    }
   
    // should throw an error
    public static CallResult GetResultStructBoolFfiError()
    {
        return new CallResult(get_result_struct_bool_ffi_error());
    }

    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern CallResultInternal get_result_struct_bool_ffi_success();
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern CallResultInternal get_result_struct_bool_ffi_error();

    // should return the result type of FakeResult
    public static FakeResultString GetResultStructStringFfiSuccess()
    {
        var result = new CallResult(get_result_struct_string_ffi_success());
        return Marshal.PtrToStructure<FakeResultString>(result.ResultPtr());
    }
   
    // should throw an error
    public static CallResult GetResultStructStringFfiError()
    {
        return new CallResult(get_result_struct_string_ffi_error());
    }

    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern CallResultInternal get_result_struct_string_ffi_success();
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern CallResultInternal get_result_struct_string_ffi_error();
    
    public static string PtrToString(IntPtr ptr)
    {
        var str = Marshal.PtrToStringUTF8(ptr);
        free_str(ptr);
        return str;
    }
    
    public static string PtrToStringNoFree(IntPtr ptr)
    {
        return Marshal.PtrToStringUTF8(ptr);
    }
    
    [DllImport(FFI.NativeLib, CallingConvention = CallingConvention.Cdecl)]
    public static extern void free_str(IntPtr s);
}



public static class ABIS { 
    public static string ERC20 = @"
        {
          ""abi"": [
            {
              ""anonymous"": false,
              ""inputs"": [
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""spender"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": false,
                  ""internalType"": ""uint256"",
                  ""name"": ""value"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""Approval"",
              ""type"": ""event""
            },
            {
              ""anonymous"": false,
              ""inputs"": [
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": false,
                  ""internalType"": ""uint256"",
                  ""name"": ""value"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""Transfer"",
              ""type"": ""event""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""spender"",
                  ""type"": ""address""
                }
              ],
              ""name"": ""allowance"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""spender"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""amount"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""approve"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""account"",
                  ""type"": ""address""
                }
              ],
              ""name"": ""balanceOf"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""decimals"",
              ""outputs"": [
                {
                  ""internalType"": ""uint8"",
                  ""name"": """",
                  ""type"": ""uint8""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""spender"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""subtractedValue"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""decreaseAllowance"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""spender"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""addedValue"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""increaseAllowance"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""name"",
              ""outputs"": [
                {
                  ""internalType"": ""string"",
                  ""name"": """",
                  ""type"": ""string""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""symbol"",
              ""outputs"": [
                {
                  ""internalType"": ""string"",
                  ""name"": """",
                  ""type"": ""string""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""totalSupply"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""amount"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""transfer"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""amount"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""transferFrom"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            }
          ]
        }
    ";

    public static string ERC721 = @"
        {
          ""abi"": [
            {
              ""inputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""constructor""
            },
            {
              ""anonymous"": false,
              ""inputs"": [
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""approved"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""Approval"",
              ""type"": ""event""
            },
            {
              ""anonymous"": false,
              ""inputs"": [
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""operator"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": false,
                  ""internalType"": ""bool"",
                  ""name"": ""approved"",
                  ""type"": ""bool""
                }
              ],
              ""name"": ""ApprovalForAll"",
              ""type"": ""event""
            },
            {
              ""anonymous"": false,
              ""inputs"": [
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""indexed"": true,
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""Transfer"",
              ""type"": ""event""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""approve"",
              ""outputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                }
              ],
              ""name"": ""balanceOf"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""getApproved"",
              ""outputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": """",
                  ""type"": ""address""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""operator"",
                  ""type"": ""address""
                }
              ],
              ""name"": ""isApprovedForAll"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""name"",
              ""outputs"": [
                {
                  ""internalType"": ""string"",
                  ""name"": """",
                  ""type"": ""string""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""ownerOf"",
              ""outputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": """",
                  ""type"": ""address""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""safeTransferFrom"",
              ""outputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                },
                {
                  ""internalType"": ""bytes"",
                  ""name"": ""data"",
                  ""type"": ""bytes""
                }
              ],
              ""name"": ""safeTransferFrom"",
              ""outputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""operator"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""bool"",
                  ""name"": ""approved"",
                  ""type"": ""bool""
                }
              ],
              ""name"": ""setApprovalForAll"",
              ""outputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""bytes4"",
                  ""name"": ""interfaceId"",
                  ""type"": ""bytes4""
                }
              ],
              ""name"": ""supportsInterface"",
              ""outputs"": [
                {
                  ""internalType"": ""bool"",
                  ""name"": """",
                  ""type"": ""bool""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""symbol"",
              ""outputs"": [
                {
                  ""internalType"": ""string"",
                  ""name"": """",
                  ""type"": ""string""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""index"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""tokenByIndex"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""owner"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""index"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""tokenOfOwnerByIndex"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""tokenURI"",
              ""outputs"": [
                {
                  ""internalType"": ""string"",
                  ""name"": """",
                  ""type"": ""string""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [],
              ""name"": ""totalSupply"",
              ""outputs"": [
                {
                  ""internalType"": ""uint256"",
                  ""name"": """",
                  ""type"": ""uint256""
                }
              ],
              ""stateMutability"": ""view"",
              ""type"": ""function""
            },
            {
              ""inputs"": [
                {
                  ""internalType"": ""address"",
                  ""name"": ""from"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""address"",
                  ""name"": ""to"",
                  ""type"": ""address""
                },
                {
                  ""internalType"": ""uint256"",
                  ""name"": ""tokenId"",
                  ""type"": ""uint256""
                }
              ],
              ""name"": ""transferFrom"",
              ""outputs"": [],
              ""stateMutability"": ""nonpayable"",
              ""type"": ""function""
            }
          ]
        }
    ";
}